""" Fixtures that the package provides.

This module provides 3 main fixtures:
:func:`~pytest_flask_socketio.fixtures.client`,
:func:`~pytest_flask_socketio.fixtures.clients`, and
:func:`~pytest_flask_socketio.fixtures.client_factory`

For each fixture you can specify a clients_receiver object, which is an instance
of :class:`EventReceiver <pytest_flask_socketio.fixtures.EventReceiver>`,
in a variable.

You can specify variables for fixtures in any of marks, class variables,
module variables, or leave it to the default. In the case of multiple, the
closest thing to the test will be given preference::

    # test_app.py

    num_clients = 3

    class TestSomething:
        num_clients = 4

        @pytest.mark.socketio(num_clients=8)
        def test_a(clients):
            print('should be 8', len(clients))

        def test_b(clients):
            print('should be 4', len(clients))

    def test_it(clients):
        print('should be 3', len(clients))
"""

import pytest
from flask_socketio import SocketIOTestClient

from pytest_flask_socketio.utils import make_clients


@pytest.fixture
def _event_receiver(request, pytestconfig):
    return _get_val(request, pytestconfig, 'event_receiver',
                    default=DefaultEventReceiver)


@pytest.fixture
def client_factory(app, socketio, _event_receiver, request, pytestconfig):
    """ Fixture for a callable to create a client.

    This fixture might be useful if you want to create an arbitrary number of
    clients on the fly::

        def test_something(client_factory):
            for _ in range(10):
                client = client_factory()
                client.emit('something')

    :return: A callable object to get a client.
    """
    if pytestconfig.flask_socketio_automark:
        request.applymarker('socketio')
    return lambda: PytestSocketIOTestClient(
        app=app,
        socketio=socketio,
        flask_test_client=app.test_client(),
        event_receiver=_event_receiver
    )


@pytest.fixture
def client(client_factory, pytestconfig):
    """ Fixture for a single client.

    This just calls the client factory one time for creating a single test
    client::

        def test_something(client):
            assert client.is_connected()
    """
    return client_factory()


@pytest.fixture
def clients(client_factory, request, pytestconfig):
    """ Fixture for many clients.

    This will create many clients and provide the test function with either a
    list of test clients (default), or a dictionary. You can use a
    ``num_clients`` variable to set the number of clients. To change the clients
    object to be a dictionary, you can set ``clients_fmt="alpha"`` to use
    letters as the identifiers (or ``"alpha_upper"`` for uppercase), or just
    ``"dict"`` to use numbers as identifiers::

        clients_fmt = 'alpha_upper'
        num_clients = 8

        def test_something(clients):
            assert clients['A'].is_connected()
            assert clients['B'].is_connected()
            assert clients['C'].is_connected()
            assert clients['F'].is_connected()

    You can also set the ``client_idents`` to be a custom list::

        client_idents = ['Bob', 'Joe']

        def test_something(clients):
            assert clients['Bob'].is_connected()
            assert clients['Joe'].is_connected()
    """
    num_clients = _get_val(request, pytestconfig, name='num_clients')
    clients_fmt = _get_val(request, pytestconfig, name='clients_fmt')
    client_idents = _get_val(request, pytestconfig, name='client_idents')
    num_clients = num_clients if client_idents is None else len(client_idents)
    return make_clients(
        client_factory=client_factory,
        num_clients=num_clients,
        idents=client_idents,
        fmt=clients_fmt,
    )


# ============================================================================ #

# Helper which first checks for a socketio marker to specify the argument, then
# looks for a member of the resulting class, module, and session (in that
# order), then the config. If not specified anywhere returns default kwarg.
def _get_val(request, pytestconfig, name, default=None):
    marker = request.node.get_closest_marker('socketio')
    if marker is not None and name in marker.kwargs:
        return marker.kwargs[name]

    for o in [request.cls, request.module, request.session, pytestconfig]:
        attr = f'flask_socketio_{name}' if o is pytestconfig else name
        if o is not None and hasattr(o, attr):
            return getattr(o, attr)
    return default


# ============================================================================ #


class EventReceiver(list):
    """ A basic class that wraps the list of received data.

    You could write a custom class to wrap the list from
    ``SocketIOTestClient.get_received()`` and set it to a ``event_receiver``
    variable in your module/class::

       class MyReceiver(EventReceiver):
           def has_error(self, i=0) -> bool:
               return self[i]['name'] == 'error'

       event_receiver = MyReceiver

    """
    def __init__(self, data):
        list.__init__(self, data)
        self.data = data


class DefaultEventReceiver(EventReceiver):
    """ Default implementation with some simple helpers.

    You could also subclass this to decorate it and add more helpers::

       class MyReceiver(DefaultEventReceiver):
           def has_error(self, i=0) -> bool:
               return self.has_message('error', i)

       event_receiver = MyReceiver

    """

    def has_message(self, msg, i=0) -> bool:
        return self[i]['name'] == msg

    def has_arg(self, key, i=0, argsi=0) -> bool:
        return self.get_from_arg(key, i, argsi) is not None

    def get_from_arg(self, key, i=0, argsi=0) -> bool:
        return self[i]['args'][argsi][key] if key in self[i]['args'][argsi] \
            else None


class PytestSocketIOTestClient(SocketIOTestClient):
    def __init__(self, app, socketio, namespace=None, query_string=None,
                 headers=None, flask_test_client=None, event_receiver=None):
        super().__init__(app, socketio, namespace, query_string,
                         headers, flask_test_client)
        self.event_receiver = event_receiver

    def get_received(self, namespace=None):
        received = super().get_received(namespace)
        return received if self.event_receiver is None \
            else self.event_receiver(received)
