import string
from typing import List, Union, Dict


idents_iterable_map = {
    'alpha': string.ascii_lowercase,
    'alpha_upper': string.ascii_uppercase,
}


def make_clients(
        client_factory: callable,
        num_clients: int = 5,
        fmt: str = 'list',
        idents: List[str] = None,
        *args, **kwargs
) -> Union[List, Dict]:
    """ Function to make several instances of test clients.

    :param client_factory: A callable to create a new test client.
    :param num_clients: The number of clients to create (default 5). Is ignored
                        if a custom value is supplied to idents in the call.
    :param fmt: One of ``'list'``, ``'num'``, ``'alpha'`` or ``'alpha_upper'``
                - is ignored if idents is provided. ``'list'`` returns a list,
                ``'alpha'`` a dictionary with lower-case letters as the key,
                ``'alpha_upper'`` for upper, and ``'num'`` for numerically
                identified string keys.
    :param idents: A custom list of identifiers to use for the dictionary.
    :param args: Positional arguments to pass to the client factory.
    :param kwargs: Keyword arguments to pass to the client factory.
    :returns: A list or dict of test clients produced by ``client_factory``.
    """

    if idents is None:
        idents = idents_iterable_map.get(fmt, range(num_clients))
    else:
        fmt = 'custom'

    if num_clients > len(idents):
        raise TypeError('cannot have more clients than identifiers')

    clients = {}
    for _, ident in zip(range(num_clients), idents):
        clients[str(ident)] = client_factory(*args, **kwargs)

    return list(clients.values()) if fmt == 'list' else clients
